#include "input.h"
#include "main.h"

void Set_View_Transformation_Matrix(view *View,
                                    double **View_Transformation_Matrix)
{
/*
 * Calculation based on Mathematical Elements for Computer Graphics
 * Page 123
 */
   eye            *Eye;
   int             Index_Eye;
   double          Sum;
   int             Translation_Matrix_Dirty;
   int             xy_Rotation_Matrix_Dirty;
   int             z_Rotation_Matrix_Dirty;
   static double **Translation_Matrix = NULL;
   static double **xy_Rotation_Matrix = NULL;
   static double **z_Rotation_Matrix = NULL;
   static double **Prespective_Matrix = NULL;
   static double **Screen_Matrix = NULL;
   static double **Transformation_Matrix = NULL;
   static double **Temp = NULL;
   double          New_Up[3];
   double          d;

   if (Translation_Matrix == NULL)
      Translation_Matrix = Get_Transformation_Matrix();

   if (xy_Rotation_Matrix == NULL)
      xy_Rotation_Matrix = Get_Transformation_Matrix();

   if (z_Rotation_Matrix == NULL)
      z_Rotation_Matrix = Get_Transformation_Matrix();

   if (Prespective_Matrix == NULL)
      Prespective_Matrix = Get_Transformation_Matrix();

   if (Screen_Matrix == NULL)
      Screen_Matrix = Get_Transformation_Matrix();

   if (Transformation_Matrix == NULL)
      Transformation_Matrix = Get_Transformation_Matrix();

   if (Temp == NULL)
      Temp = Get_Transformation_Matrix();

   Eye = View->Eye;
   Index_Eye = View->Current_Eye_Point;

/*
 *  Set the view point at origin
 */

   SET_ID_MATRIX(Translation_Matrix);

   Sum = Translation_Matrix[3][X] = -(Eye[Index_Eye].Location[X]);
   Sum += Translation_Matrix[3][Y] = -(Eye[Index_Eye].Location[Y]);
   Sum += Translation_Matrix[3][Z] = -(Eye[Index_Eye].Location[Z]);

   if (fabs(Sum) >= SMALL_NUMBER)
      Translation_Matrix_Dirty = TRUE;
   else
      Translation_Matrix_Dirty = FALSE;

#ifdef DEBUG_Set_Transformation_Matrix
   PRINT_MATRIX("Translation Matrix", Translation_Matrix);
#endif

/*
 * Rotate the view plane normal about the x-axis.
 */

   d = sqrt((Eye[Index_Eye].View_direction[J_HAT]
         * Eye[Index_Eye].View_direction[J_HAT])
      + (Eye[Index_Eye].View_direction[K_HAT]
         * Eye[Index_Eye].View_direction[K_HAT]));

   xy_Rotation_Matrix_Dirty = FALSE;

   if ((d > SMALL_NUMBER)
       && ((1.0 - Eye[Index_Eye].View_direction[K_HAT]) > SMALL_NUMBER))
   {
      SET_ID_MATRIX(xy_Rotation_Matrix);

      xy_Rotation_Matrix_Dirty = TRUE;

#ifdef DEBUG_Set_Transformation_Matrix
      printf("rotate about z axis\n");
#endif

      xy_Rotation_Matrix[1][J_HAT] = xy_Rotation_Matrix[2][K_HAT]
         = Eye[Index_Eye].View_direction[K_HAT] / d;
      xy_Rotation_Matrix[1][K_HAT]
         = Eye[Index_Eye].View_direction[J_HAT] / d;
      xy_Rotation_Matrix[2][J_HAT] = -(xy_Rotation_Matrix[1][K_HAT]);

#ifdef DEBUG_Set_Transformation_Matrix
      PRINT_MATRIX("xy_Rotation Matrix (x)", xy_Rotation_Matrix);
#endif
   }

/*
 * Rotate the view plane normal about the y-axis.
 */

   if (fabs(Eye[Index_Eye].View_direction[I_HAT]) > SMALL_NUMBER)
   {
      if (xy_Rotation_Matrix_Dirty == FALSE)
      {
         xy_Rotation_Matrix_Dirty = TRUE;
         SET_ID_MATRIX(xy_Rotation_Matrix);

         xy_Rotation_Matrix[0][I_HAT] = xy_Rotation_Matrix[2][K_HAT] = d;
         xy_Rotation_Matrix[0][K_HAT]
            = Eye[Index_Eye].View_direction[I_HAT];
         xy_Rotation_Matrix[2][I_HAT] = -(xy_Rotation_Matrix[0][K_HAT]);
      }
      else
      {
         SET_ID_MATRIX(Temp);

#ifdef DEBUG_Set_Transformation_Matrix
         printf("rotate about y axis\n");
#endif

         Temp[0][I_HAT] = Temp[2][K_HAT] = d;
         Temp[0][K_HAT] = Eye[Index_Eye].View_direction[I_HAT];
         Temp[2][I_HAT] = -(Temp[0][K_HAT]);

#ifdef DEBUG_Set_Transformation_Matrix
         PRINT_MATRIX("Rotation Matrix (y)", Temp);
#endif

         Matrix_4x4_Multiplication(xy_Rotation_Matrix, Temp,
            xy_Rotation_Matrix);
      }
   }

/*
 * Transform the the view plane up vector to the new coordinate system.
 */

   if (xy_Rotation_Matrix_Dirty)
   {
      Vector_Matrix_Multiplication(Eye[Index_Eye].Up,
         xy_Rotation_Matrix, New_Up);
   }
   else
   {
      New_Up[X] = Eye[Index_Eye].Up[X];
      New_Up[Y] = Eye[Index_Eye].Up[Y];
      New_Up[Z] = Eye[Index_Eye].Up[Z];
   }

#ifdef DEBUG_Set_Transformation_Matrix
   printf("New_Up = (%g, %g, %g)\n", New_Up[I_HAT],
      New_Up[J_HAT], New_Up[K_HAT]);
#endif

   if (fabs(New_Up[I_HAT]) > SMALL_NUMBER)
   {

#ifdef DEBUG_Set_Transformation_Matrix
      double          Test_Vector[3];

      printf("rotate about z axis\n");
#endif

/*
 * The up vector is not pointing the same direction of y-axis.
 * Therefore, rotate the system to match the direction of y-axis.
 */
      z_Rotation_Matrix_Dirty = TRUE;

      SET_ID_MATRIX(z_Rotation_Matrix);

      z_Rotation_Matrix[0][I_HAT] = z_Rotation_Matrix[1][J_HAT]
         = New_Up[J_HAT];

      z_Rotation_Matrix[0][J_HAT] = New_Up[I_HAT];

      z_Rotation_Matrix[1][I_HAT] = -(z_Rotation_Matrix[0][J_HAT]);

#ifdef DEBUG_Set_Transformation_Matrix
      PRINT_MATRIX("Rotation Matrix (z)", z_Rotation_Matrix);

      Vector_Matrix_Multiplication(New_Up, z_Rotation_Matrix,
         Test_Vector);
      printf("Test_Vector = (%g, %g, %g)\n", Test_Vector[I_HAT],
         Test_Vector[J_HAT], Test_Vector[K_HAT]);
#endif
   }
   else
   {
      z_Rotation_Matrix_Dirty = FALSE;
   }

/*
 * Prespective transformation.
 */

   SET_ID_MATRIX(Prespective_Matrix);

   Prespective_Matrix[2][H_HAT] = 1.0 / View->Distance_to_Screen;

#ifdef DEBUG_Set_Transformation_Matrix
   PRINT_MATRIX("Prespective Matrix", Prespective_Matrix);
#endif

   if (Translation_Matrix_Dirty)
   {
      COPY_MATRIX(Transformation_Matrix, Translation_Matrix);
   }
   else
   {
      SET_ID_MATRIX(Transformation_Matrix);
   }

#ifdef DEBUG_Set_Transformation_Matrix
   PRINT_MATRIX("Transformation Matrix", Transformation_Matrix);
#endif

   if (xy_Rotation_Matrix_Dirty)
   {
      Matrix_4x4_Multiplication(Transformation_Matrix, xy_Rotation_Matrix,
         Transformation_Matrix);

#ifdef DEBUG_Set_Transformation_Matrix
      PRINT_MATRIX("Transformation Matrix", Transformation_Matrix);
#endif
   }

   if (z_Rotation_Matrix_Dirty)
   {
      Matrix_4x4_Multiplication(Transformation_Matrix, z_Rotation_Matrix,
         Transformation_Matrix);

#ifdef DEBUG_Set_Transformation_Matrix
      PRINT_MATRIX("Transformation Matrix", Transformation_Matrix);
#endif
   }

   Matrix_4x4_Multiplication(Transformation_Matrix, Prespective_Matrix,
      Transformation_Matrix);

   printf("Transformation_Matrix[0][0] = %g\n", Transformation_Matrix[0][0]);
   PRINT_MATRIX("Transformation Matrix", Transformation_Matrix);
#ifdef DEBUG_Set_Transformation_Matrix
   PRINT_MATRIX("Transformation Matrix", Transformation_Matrix);
#endif

   COPY_MATRIX(View_Transformation_Matrix, Transformation_Matrix);
}

void Set_Image_Transformation_Matrix(window *Image, view *View,
                                     double **Image_Transformation_Matrix)
{

   SET_ID_MATRIX(Image_Transformation_Matrix);

   printf("Image LL (%d, %d) UR (%d, %d)\n",
          Image -> LL_x, Image -> LL_y,
          Image -> UR_x, Image -> UR_y);

   printf("Screen LL (%d, %d) UR (%d, %d)\n",
          View -> Screen.LL_x, View -> Screen.LL_y,
          View -> Screen.UR_x, View -> Screen.UR_y);
#if 0
   Image_Transformation_Matrix[X][X]
      = (double) (Image -> UR_x - Image -> LL_x)
        / (double) (View -> Screen.UR_x - View -> Screen.LL_x);
   Image_Transformation_Matrix[Y][Y]
      = (double) (Image -> UR_y - Image -> LL_y)
        / (double) (View -> Screen.UR_y - View -> Screen.LL_y);
#endif
   Image_Transformation_Matrix[3][X] = -((double) View -> Screen.LL_x);
   Image_Transformation_Matrix[3][Y] = -((double) View -> Screen.LL_y);

   printf("Image_Transformation_Matrix[X][X] = %g\n",
          Image_Transformation_Matrix[X][X]);
   printf("Image_Transformation_Matrix[Y][Y] = %g\n",
          Image_Transformation_Matrix[Y][Y]);
   printf("Image_Transformation_Matrix[3][X] = %g\n",
          Image_Transformation_Matrix[3][X]);
   printf("Image_Transformation_Matrix[3][Y] = %g\n",
          Image_Transformation_Matrix[3][Y]);
}

